from enum import Enum
from datetime import datetime
from textual import on
from textual.app import App, ComposeResult, Binding
from textual.containers import HorizontalGroup, VerticalScroll
from textual.reactive import reactive
from textual.widgets import Button, Digits, Footer, Header, DataTable, Label, Rule, Input
from textual.color import Color

from twd.config import Config
from twd.data import TwdManager
from twd.utils import fuzzy_search, linear_search
from twd.modals import ConfirmModal, EntryDeleteModal

class Mode(Enum):
    NORMAL = "normal"
    SEARCH = "search"

class TWDApp(App):
    """
    TWD TUI Application
    """

    CSS_PATH = "tui.tcss"

    BINDINGS = [
            # motion
            Binding("j", "cursor_down", "Down"),
            Binding("k", "cursor_up", "Up"),

            # modify
            Binding("/", "enter_search_mode", "Search"),
            Binding("d", "delete_entry", "Delete"),
            Binding("escape", "enter_normal_mode", "Normal", show=False),
            # TODO: edit
            # TODO: rename

            # select
            Binding("enter", "select", "Select"),

            # exit
            Binding("q", "exit", "Exit"),
        ]

    mode: Mode = reactive(Mode.NORMAL)

    def __init__(self, manager: TwdManager, *args, **kwargs):
        self.manager = manager
        super().__init__(*args, **kwargs)

    def compose(self) -> ComposeResult:
        yield Header(show_clock=True)
        yield Footer()

        # cwd
        yield HorizontalGroup(
                Label(f"cwd: {self.manager.cwd}", classes="cwd"),
                Label("", id="mode"),
                id="header"
                )

        yield Input(placeholder="Search...", id="search-input")

        # twd selection table
        yield DataTable(
                cursor_type='row',
                cell_padding=2,
                # zebra_stripes=True,
                id="data",
                )

    def on_mount(self) -> None:
        # app config
        self.theme = "flexoki"
        self.title = "TWD"
        self.sub_title = "Tracked Working Directory"

        search_input = self.query_one("#search-input", Input)
        search_input.display = False

        # add headers
        table = self.query_one(DataTable)
        table.add_columns(*self.manager.CSV_HEADERS_FANCY)
        
        self._populate_table()

    def _populate_table(self, entries=None) -> None:
        """
        fill or refresh data table
        """
        table = self.query_one(DataTable)
        table.clear()

        if entries is None:
            entries = self.manager.list_all()
        
        # fill data
        for entry in entries:
            table.add_row(entry.alias, str(entry.path), entry.name, entry.created_at.strftime("%Y-%m-%d %H:%M:%S"))

    def watch_mode(self, old_mode: Mode, new_mode: Mode) -> None:
        """
        react to mode changes
        """
        search_input = self.query_one("#search-input", Input)
        table = self.query_one(DataTable)
        mode_label = self.query_one("#mode")

        mode_label.update(str(new_mode.value).upper())

        if new_mode == Mode.SEARCH:
            # enter search mode
            search_input.display = True
            search_input.value = ""
            search_input.focus()

            mode_label.add_class("search")
        elif new_mode == Mode.NORMAL:
            # enter normal mode
            search_input.display = False
            search_input.value = ""
            self._populate_table()
            table.focus()

            mode_label.remove_class("search")

    # actions
    def action_cursor_down(self) -> None:
        """
        move cursor down
        """
        table = self.query_one(DataTable)

        try:
            current_row = table.cursor_coordinate.row
            next_row = (current_row + 1) % table.row_count
        except ZeroDivisionError as e:
            return

        table.move_cursor(row=next_row)

    def action_cursor_up(self) -> None:
        """
        move cursor up
        """
        table = self.query_one(DataTable)
        
        try:
            current_row = table.cursor_coordinate.row
            prev_row = (current_row - 1) % table.row_count
        except ZeroDivisionError as e:
            return

        table.move_cursor(row=prev_row)

    def action_enter_search_mode(self) -> None:
        """
        enter search mode
        """
        if self.mode == Mode.SEARCH:
            return
        self.mode = Mode.SEARCH

    def action_enter_normal_mode(self) -> None:
        """
        enter normal mode
        """
        if self.mode == Mode.NORMAL:
            return
        self.mode = Mode.NORMAL

    def action_delete_entry(self) -> None:
        """
        open confirm modal and delete entry if yes
        """
        if not self.mode == Mode.NORMAL:
            return

        table = self.query_one(DataTable)

        # get row
        row_key = table.coordinate_to_cell_key(table.cursor_coordinate).row_key
        row_data = table.get_row(row_key)
        alias = row_data[0]

        # get entry
        entry = self.manager.get(alias)

        def check_delete(decision: bool | None) -> None:
            """
            open modal and return the decision
            """
            if not decision:
                return

            self.manager.remove(alias)

            self._populate_table()

            self.notify(f"Removed entry \"{entry.name}\"")

        # self.push_screen(ConfirmModal(message=f"Are you sure want to delete '{entry.alias}'?"), check_delete)
        self.push_screen(EntryDeleteModal(entry), check_delete)

    def action_exit(self) -> None:
        self.exit()

    @on(Input.Changed, "#search-input")
    def on_search_input_changed(self, e: Input.Changed) -> None:
        """
        filter table as user types
        """
        if self.mode != Mode.SEARCH:
            return

        query = e.value

        all_entries = self.manager.list_all()

        # TODO: filter entries and repopulate table

        if query is None:
            self._populate_table()
            return

        search_result = fuzzy_search(query, all_entries)

        filtered = [item[0] for item in search_result]

        self._populate_table(filtered)

    @on(Input.Submitted, "#search-input")
    def on_search_submitted(self, e: Input.Submitted) -> None:
        """
        when user presses enter in search, return to normal mode
        """
        if self.mode != Mode.SEARCH:
            return

        self.mode = Mode.NORMAL

        self.query_one(DataTable).focus()


    @on(DataTable.RowSelected)
    def on_data_table_row_selected(self, event: DataTable.RowSelected) -> None:
        """
        select row and send back to the original program (probably cli)
        """
        table = event.data_table
        row_key = event.row_key

        # get row
        row_data = table.get_row(row_key)
        alias = row_data[0]

        # get entry
        entry = self.manager.get(alias)

        self.notify(f"Selected: {entry.alias} -> {entry.path}")

        # return selected path to cli
        self.exit(entry.path) 

if __name__ == "__main__":
    # made sure it works with 'serve'
    config = Config.load()
    manager = TwdManager(config.data_path)

    app = TWDApp(manager=manager)
    path = app.run()

    print(path)



import click
import os
from pathlib import Path
from . import __version__
from .config import Config
from .data import TwdManager
from .tui import TWDApp

@click.group(invoke_without_command=True)
@click.pass_context
@click.version_option(version=__version__)
def cli(ctx):
    """
    TWD - Temp / Tracked Working Directory
    """
    ctx.ensure_object(dict)

    ctx.obj['config'] = Config.load() # load config
    ctx.obj['manager'] = TwdManager(ctx.obj['config'].data_path)

    # if no subcommand was provided, launch TUI
    if ctx.invoked_subcommand is None:
        # TODO: launch TUI here

        path = TWDApp(manager=ctx.obj['manager']).run()

        if not path:
            print("Exiting...")
            exit(0)

        # write to fd3
        os.write(3, bytes(str(path), "utf-8"))
        exit(0)

@cli.command()
@click.argument('path')
@click.argument('alias', required=False)
@click.argument('name', required=False)
@click.pass_context
def save(ctx, path, alias, name):
    """
    Save a new twd with PATH, [opt] ALIAS and [opt] NAME
    """
    manager: TwdManager = ctx.obj['manager']

    path_obj = Path(path).expanduser().resolve()

    # get alias from path
    if alias is None:
        alias = path_obj.name.lower().replace(" ", "_")

    try:
        entry = manager.add(alias, path_obj, name)
        click.echo(f"Saved '{entry.alias}' -> {entry.path}")
    except ValueError as e:
        click.echo(f"Error: {e}")
        raise click.Abort()

@cli.command()
@click.argument('alias')
@click.pass_context
def get(ctx, alias):
    """
    get path for ALIAS
    """
    manager = ctx.obj['manager']

    entry = manager.get(alias)

    if entry is None:
        click.echo(f"Alias '{alias}' not found")
        raise click.Abort()

    # TODO: os.write(3) / write to fd3 for shell integration
    os.write(3, bytes(str(entry.path), "utf-8"))

    click.echo(f"cd-ing to {entry.path}")

@cli.command()
@click.argument('alias')
@click.pass_context
def remove(ctx, alias):
    """
    remove TWD by ALIAS
    """
    manager = ctx.obj['manager']
    
    try:
        manager.remove(alias)
        click.echo(f"Removed '{alias}'")
    except KeyError as e:
        click.echo(f"{e}", err=True)
        raise click.Abort()

@cli.command('list')
@click.pass_context
def list_twds(ctx):
    """List all TWDs"""
    manager = ctx.obj['manager']
    
    entries = manager.list_all()
    if not entries:
        click.echo("No TWDs saved yet. Use 'twd save <path> <alias>' to add one.")
        return
    
    invalid_found = False
    for entry in entries:
        if not invalid_found and not os.path.exists(entry.path):
            invalid_found = True
        click.echo(f"{entry.alias:20} {entry.name:30} {entry.path}")

    if invalid_found:
        click.echo(f"\nInvalid TWD paths found. Run <twd clean> to remove them.")


@cli.command('clean')
@click.option('--yes', '-y', is_flag=True, help="Remove all invalid entries without asking")
@click.pass_context
def clean(ctx, yes):
    """
    clean twds from invalid paths
    """
    manager = ctx.obj['manager']

    # all entries, valid and invalid
    entries = manager.list_all()

    # only invalids
    invalids = []
    for entry in entries:
        if os.path.exists(entry.path):
            continue

        invalids.append(entry)

    if len(invalids) == 0:
        click.echo("No invalid TWDs found.")
        return

    click.echo(f"Found {len(invalids)} invalid TWDs\n")

    # only valid
    valid_entries = []
    for entry in entries:
        if entry not in invalids:
            valid_entries.append(entry)

    # remove all
    if yes:
        click.echo("Removing all invalid entries...")
        manager._write_all(valid_entries)
        click.echo(f"Done. {len(valid_entries)} TWDs left.")
        return

    to_keep = []
    for inv in invalids:
        if not click.confirm(f"Do you want to remove '{inv.alias}'?", default=True):
            # user wants to keep invalid TWD (weird but i'll allow it)
            to_keep.append(inv)

    # write a unison list of valid entries and the ones to keep
    final_entries = valid_entries + to_keep
    manager._write_all(final_entries)
    click.echo(f"Done. {len(final_entries)} TWDs left.")



from textual import on
from textual.app import ComposeResult
from textual.screen import ModalScreen
from textual.containers import Container, Horizontal
from textual.widgets import Button, Label
from typing import Union

from twd.data import Entry

class ConfirmModal(ModalScreen[bool]):
    """A confirm modal"""

    DEFAULT_CSS = """
        ConfirmModal {
            align: center middle;
        }

        ConfirmModal > Container {
            width: auto;
            height: auto;
            border: thick $background 80%;
            background: $surface;
        }

        ConfirmModal > Container > Label {
            width: 100%;
            content-align-horizontal: center;
            margin-top: 1;
        }

        ConfirmModal > Container > Horizontal {
            width: auto;
            height: auto;
        }

        ConfirmModal > Container > Horizontal > Button {
            margin: 2 4;
        }
    """

    def __init__(
            self, 
            message: Union[str, None] = None,
            confirm_text: str = "Yes",
            cancel_text: str = "No"
    ):
        """
        message: The message to display when popping the modal
        confirm_text: Text to show for confirmation
        cancel_text: Text to show for cancellation
        """
        self.message = message
        self.confirm_text = confirm_text
        self.cancel_text = cancel_text

        super().__init__()

    def compose_content(self) -> ComposeResult:
        """
        Abstract method for presenting custom shenanigans
        """
        if self.message:
            yield Label(self.message, id="content")
        else:
            yield Label("Are you sure?", id="content")

    def compose(self) -> ComposeResult:
        with Container():
            yield from self.compose_content()
            with Horizontal():
                yield Button(self.cancel_text, id="no", variant="error")
                yield Button(self.confirm_text, id="yes", variant="success")

    @on(Button.Pressed, "#no")
    def no_decision(self) -> None:
        """decision no"""
        self.dismiss(False)

    @on(Button.Pressed, "#yes")
    def yes_decision(self) -> None:
        """decision yes"""
        self.dismiss(True)

class EntryDeleteModal(ConfirmModal):
    """Confirmation modal with detailed entry information"""

    def __init__(self, entry):
        self.entry = entry
        super().__init__(confirm_text="Delete", cancel_text="Cancel")

    def compose_content(self) -> ComposeResult:
        yield Label(f"Delete entry '{self.entry.name}'?", id="content")



import csv

from pathlib import Path
from pydantic import BaseModel, Field, validator
from typing import List, Optional
from datetime import datetime

from .config import Config

class Entry(BaseModel):
    """Data class for a signle TWD Entry"""
    
    alias: str = Field(..., min_length=2, max_length=64)
    path: Path
    name: str = Field(..., min_length=3)
    created_at: datetime = Field(default_factory=datetime.now)

    @validator("alias")
    def validate_alias(cls, v):
        if not v.replace("_", "").replace("-", "").isalnum():
            raise ValueError("Alias must be alphanumeric with - or _")

        return v.lower()

    @validator("path")
    def validate_path(cls, v):
        path = Path(v).expanduser().resolve()

        return path

    def to_csv(self) -> List[str]:
        """Convert to csv row"""
        return [
                self.alias,
                str(self.path),
                self.name,
                self.created_at.isoformat()
                ]

    @classmethod
    def from_csv(cls, row: List[str]) -> "Entry":
        """create from csv row"""
        return cls(
                alias=row[0],
                path=Path(row[1]),
                name=row[2],
                created_at=datetime.fromisoformat(row[3])
                )

class TwdManager:
    """twd entry manager stored in csv"""

    CSV_HEADERS = ["alias", "path", "name", "created_at"]
    CSV_HEADERS_FANCY = ["Alias", "Path", "Description", "Created at"]

    def __init__(self, csv_path: Path):
        self.csv_path = csv_path
        self._ensure_csv_exists()
        self.cwd = str(Path.cwd())

    def _ensure_csv_exists(self) -> None:
        """create csv headers"""
        if not self.csv_path.exists():
            self.csv_path.parent.mkdir(parents=True, exist_ok=True) 

            with open(self.csv_path, 'w', newline='') as f:
                writer = csv.writer(f)
                writer.writerow(self.CSV_HEADERS)

    def _read_all(self) -> List[Entry]:
        """read all entries"""
        entries = []

        with open(self.csv_path, 'r', newline='') as f:
            reader = csv.reader(f)

            next(reader) # skip headers

            for row in reader:
                entries.append(Entry.from_csv(row))

        return entries

    def _write_all(self, entries: List[Entry]) -> None:
        with open(self.csv_path, 'w', newline='') as f:
            writer = csv.writer(f)
            writer.writerow(self.CSV_HEADERS)

            for entry in entries:
                writer.writerow(entry.to_csv())

    def add(self, alias: str, path: Path, name: Optional[str] = None) -> Entry:
        """Add new entry"""
        entries = self._read_all()

        if any(e.alias == alias.lower() for e in entries):
            raise ValueError(f"Alias '{alias}' already exists")

        if name is None:
            name = Path(path).name

        entry = Entry(alias=alias, path=path, name=name)
        entries.append(entry)
        self._write_all(entries)

        return entry

    def get(self, alias: str) -> Optional[Entry]:
        """get entry by alias"""
        entries = self._read_all()

        for entry in entries:
            if entry.alias == alias.lower():
                return entry

        return None

    def remove(self, alias: str) -> None:
        """remove entry by alias"""
        entries = self._read_all()
        original_len = len(entries)

        entries = [e for e in entries if e.alias != alias.lower()]

        if len(entries) == original_len:
            raise KeyError(f"Alias '{alias}' not found")

        self._write_all(entries)

    def list_all(self) -> List[Entry]:
        entries = self._read_all()

        return sorted(entries, key=lambda e: e.created_at)

    def exists(self, alias: str) -> bool:
        return self.get(alias) is not None
